import os
import sys
import shutil
import subprocess
from itertools import product
from importlib.resources import files as imp_files
from mpi4py import MPI
from mpi4py.futures import MPIPoolExecutor
import logging
logger = logging.getLogger(__name__)
logger.addHandler(logging.NullHandler())

from .utils import *

class Projectors(object):
    def __init__(self, multiplets, path):
        self.mults = multiplets
        if os.path.exists(path):
            print("{0} folder exists...".format(path))
            if os.access(path, os.W_OK | os.X_OK):
                if any(os.scandir(path)):
                    user_ans=input("Directory is not empty. The content will be overwritten. Proceed (y/[n])? ")
                    if user_ans.lower() == "y":
                        for root, dirs, files in os.walk(path):
                            for f in files:
                                os.unlink(os.path.join(root, f))
                            for d in dirs:
                                shutil.rmtree(os.path.join(root, d))
                    else:
                        print("Exiting...")
                        sys.exit(0)
                self.path = path
            else:
                raise PermissionError("No write access for {0} directory!".format(path))
        else:
            raise FileNotFoundError("{0} directory does not exist!".format(path))
        #Copy SUn.prc to the output dir
        shutil.copy(imp_files('OrthoBase.FORMData').joinpath('SUn.prc'),self.path)
        print("########")

    @property
    def nodes(self):
        return(self._nodes)

    @nodes.setter
    def nodes(self, value):
        if not isinstance(value,list):
            raise TypeError("nodes must be a list of strings!")
        self._nodes = value

    @property
    def parallel_evaluation(self):
        return(self._parallel_evaluation)

    @parallel_evaluation.setter
    def parallel_evaluation(self,value):
        if not isinstance(value, bool):
            raise TypeError("parallel_evaluation must hold a boolean value!")
        self._parallel_evaluation = value

    @property
    def FORM_np(self):
        return(self._FORM_np)

    @FORM_np.setter
    def FORM_np(self,value):
        if not isinstance(value, int):
            raise TypeError("FORM_np must be an integer number of processes!")
        self._FORM_np = value

    def run(self):
        logger.info("Building expressions of projectors")
        f_old = open(os.path.join(self.path,"old_projs.frm"), "w")
        f_old_list = open(os.path.join(self.path,"old_projs_list.frm"), "w")
        f_new = open(os.path.join(self.path,"new_projs.frm"), "w")
        f_new_list = open(os.path.join(self.path,"new_projs_list.frm"), "w")
        f_new.write(imp_files('OrthoBase.FORMData').joinpath('header.frm').read_text())
        f_new.write('\n')
        f_old.write(imp_files('OrthoBase.FORMData').joinpath('header.frm').read_text())
        f_old.write('\n')
        oldproj_names = "CF "
        clebsch_names = "CF "
        clebsches = set()
        for m in self.mults:
            m.decompose()
            if self.mults.gen==m.first_occ:
                proj, oldproj = self.proj_new(m)
                f_new_list.write(proj)
                oldproj_names += oldproj + ","
            else:
                clebsch,rhs_proj_name,projs = self.proj_old(m)
                f_old_list.write("\n".join(projs))
                f_old_list.write("****************\n")
                clebsches.add(clebsch)
                clebsches.add(rhs_proj_name)
        clebsch_names += ",".join(clebsches)
        f_old.write(clebsch_names+";\n\n")
        f_old.write("#include old_projs_list.frm\n\n")
        f_new.write(oldproj_names[:-1]+";\n\n")
        f_new.write("#include new_projs_list.frm\n\n")
        #f_new.write("L test = SUNT;")
        f_new.write(imp_files('OrthoBase.FORMData').joinpath('footer_new.frm').read_text())
        f_old.write(imp_files('OrthoBase.FORMData').joinpath('footer_old.frm').read_text())
        f_new_list.close()
        f_new.close()
        f_old.close()
        f_old_list.close()

        logger.info("Starting to run FORM")
        os.chdir(self.path)
        process=subprocess.Popen(["tform",f"-w{self.FORM_np}","new_projs.frm"], stdout=subprocess.PIPE, stderr=subprocess.PIPE, text=True, bufsize=1)
        print("Program output:")
        stdout_lines = []
        while True:
            line = process.stdout.readline()
            if not line:
                break
            stdout_lines.append(line)
        process.wait()

        stdout = "".join(stdout_lines)

        print(stdout)


    def yng_operator(self, y, i):
        """
        Construct a non-hermitian Young operator
        """
        #print("Projecting ",y.dim,y.part_cols,y.part_rows,y.table)
        logger.debug(f"Projecting {y.dim},{y.part_cols},{y.part_rows},{y.table}")
        code = ""
        #symmetric combinations
        #lhs contains the indices going inside the diagram
        #rhs contains the indices coming out of the diagram
        for row,cols in enumerate(y.part_cols):
            lhs = "perm_(f1,"
            rhs = "f2("
            idx = sum(y.part_cols[:row])
            for col in range(cols):
                idx += 1
                #symm += "d(j{0},mj{0}),".format(idx)
                lhs += "{0}{1},".format(i,idx)
                rhs += "m{0}{1},".format(i,idx)
            lhs = lhs[:-1]
            rhs = rhs[:-1]
            lhs += ")"
            rhs += ")"
            code += "{0}*{1}*".format(lhs,rhs)
        #asymmetric combinations
        for col,rows in enumerate(y.part_rows):
            lhs = "perm_(1,f1,"
            rhs = "f2("
            for row in range(rows):
                idx = sum(y.part_cols[:row])+col+1
                #asymm += "d(mj{0},jj{0}),".format(idx)
                lhs += "m{0}{1},".format(i,idx)
                rhs += "{0}{0}{1},".format(i,idx)
            lhs = lhs[:-1]
            rhs = rhs[:-1]
            lhs += ")"
            rhs += ")"
            code += "{0}*{1}*".format(lhs,rhs)
        code = code[:-1]
        return code


    def proj_new(self, m):
        """
        Construction of the projectors which appear for the first time in the
        creation history

        Notation:
        i indices - quarks
        j indices - antiquarks
                             Y
                           /  \
                          /i1  \ii1
        mu1 -- P -- mmu1--       ------ mnu1 -- P -- nu1
         .                \j1   /jj1                  .
         .                 \   /                      .
         .                   Yb                       .
         .                                            .
         .               /       \                    .
        mu.... ---------   ......  ----------------- nu...
                         \       /
        """
        y_qb = m.decomposition[0]
        y_q = m.decomposition[1]
        if y_qb.ncells != m.first_occ:
            y_qb = y_qb.conjugate()
        elif y_q.ncells != m.first_occ:
            y_q = y_q.conjugate()
        print("New proj {0}={1}x{2} decomposed as {3}x{4}".format(m.dim_txt,m.parent1.dim_txt,8,y_qb.dim,y_q.dim))
        m.print()
        #parents=",".join(obj.dim_txt for obj in reversed((lambda x: [x] + (lambda x: [x] + [x.parent1])(x.parent1) if x.parent1 else [x])(m)) if obj)
        parents=",".join(obj.dim_txt for obj in reversed(m.parent_list()))+","+m.dim_txt
        proj_name = "[P({0})]".format(parents)
        all_indices = "("
        #oldproj1 = "P[{0}](".format(m.parent1.dim)
        oldproj_names = "CF "
        oldproj = "[P({0})]".format(parents[:parents.rindex(',')])
        oldproj1 = oldproj + "("
        colfac1 = ""
        for g in range(1,self.mults.gen):
            all_indices += "mu{0},".format(g)
            all_indices += "nu{0},".format(g)
            oldproj1 += "mu{0},".format(g)
            oldproj1 += "mmu{0},".format(g)
            colfac1 += "SUNT(mmu{0},i{0},j{0})*".format(g)
        all_indices += "mu{0},".format(self.mults.gen)
        all_indices += "nu{0},".format(self.mults.gen)
        all_indices = all_indices[:-1]
        all_indices += ")"
        oldproj1 = oldproj1[:-1]
        oldproj1 += ")"
        oldproj2 = oldproj1.replace("mu","nu")
        colfac1 += "SUNT(mu{0},i{0},j{0})".format(self.mults.gen)
        colfac2 = colfac1.replace("mu","nu")
        colfac2 = colfac2.replace("i","ri")
        colfac2 = colfac2.replace("j","rj")
        colfac2 = colfac2.replace("ri","jj")
        colfac2 = colfac2.replace("rj","ii")
        y_q_code=self.yng_operator(y_q, "i")
        y_qb_code=self.yng_operator(y_qb, "j")
        #Subtract the rest
        subtr = "("
        for pM in m.decomposition[2]:
            subtr += "[P({0})]".format(parents[:parents.rindex(',')]+","+str(pM.dim))
            oldproj += ",[P({0})]".format(parents[:parents.rindex(',')]+","+str(pM.dim))
            subtr += f"{all_indices}+"
        subtr = subtr[:-1]
        subtr += ")"
        proj_name += all_indices
        proj = "L {0}={1}*{2}*{3}*{4}*{5}*{6}-{7};\n".format(proj_name,oldproj1,colfac1,oldproj2,colfac2,y_q_code,y_qb_code,subtr)
        print(proj)
        return(proj,oldproj)
        #m.decomposition[1].print()
        #print(m.first_occ)

    def proj_old(self, m):
        """
        Construction of the projectors which previously already appeared in the
        creation history
        """

        print("Old proj {0}={1}x{2}".format(m.dim_txt,m.parent1.dim_txt,8))
        parents=",".join(obj.dim_txt for obj in reversed(m.parent_list()))+","+m.dim_txt
        proj_name = "[C({0})]".format(parents)
        rhs_proj_name = "[P({0})]".format(m.dim)
        clebsch = "[C({0})]".format(parents[:parents.rindex(',')])
        clebsch_idx = "("

        for g in range(1,self.mults.gen):
            clebsch_idx += "mu{0},".format(g)

        colfac1 = ""
        m.parent1.decompose()
        i_list = list()
        j_list = list()
        for g in range(1,m.parent1.first_occ+1):
            clebsch_idx += "mmu{0},".format(g)
            colfac1 += "SUNT(mmu{0},i{0},j{0})*".format(g)
            i_list.append(f"i{g}")
            j_list.append(f"j{g}")
        colfac1 += "SUNT(mu{0},i{0},j{0})".format(self.mults.gen)
        i_list.append(f"i{self.mults.gen}")
        j_list.append(f"j{self.mults.gen}")
        clebsch_idx = clebsch_idx[:-1]
        clebsch_idx += ")"

        ii_list = list()
        jj_list = list()
        colfac2 = ""
        rhs_idx = "("
        if m.first_occ==0:
            #colfac2 = "1"
            colfac2 = "d_(nu1,nu2)"
            rhs_proj_name = "1"
            rhs_idx = ""
        else:
            for g in range(1,m.first_occ+1):
                rhs_idx += "mnu{0},".format(g)
                rhs_idx += "nu{0},".format(g)
                colfac2 += "SUNT(mnu{0},jj{0},ii{0})*".format(g)
                ii_list.append(f"ii{g}")
                jj_list.append(f"jj{g}")
            colfac2 = colfac2[:-1]
            rhs_idx = rhs_idx[:-1]+")"

        perms_ij  = list(product(i_list,j_list))
        perms_iii = list(product(i_list,ii_list))
        perms_iijj= list(product(ii_list,jj_list))
        perms_jjj = list(product(j_list,jj_list))
        all_perms = perms_ij+perms_iii+perms_iijj+perms_jjj
        logger.debug(f"Indices to pair: i_list={i_list}, j_list={j_list}, ii_list={ii_list}, jj_list={jj_list}")
        pairs=generate_pairings(all_perms,len(i_list)+len(ii_list))
        results = list()
        for n,pairing in enumerate(pairs):
            proj_name_n = proj_name.replace("]", f"{n}]")
            connector='*'.join(f'd_({i},{j})' for i, j in pairing)
            proj = f"L {proj_name_n}={clebsch}{clebsch_idx}*{colfac1}*{colfac2}*{connector}*{rhs_proj_name}{rhs_idx};\n"
            results.append(proj)
            print(proj)
        if rhs_proj_name == "1":
            rhs_proj_name = ""
        return(clebsch,rhs_proj_name,results)



import z3
#import pickle
import copy
import re
from collections import Counter
import logging
logger = logging.getLogger(__name__)
logger.addHandler(logging.NullHandler())

from .utils import *
database = {}

class YoungTableaux(object):
    def __init__(self, obj, g=1):
        self.tables = obj
        dims=set()
        for sol in self.tables:
            self.gen = g+1
            dims.add(sol.dim_txt)
        self.dims = dims

    def __getitem__(self, i):
        if isinstance(i,int):
            return self.tables[i]
        elif isinstance(i,str):
            return [tab for tab in self.tables if tab.dim_txt == i]

    def __len__(self):
        return len(self.tables)

    def __mul__(self,other):
        print("Mult ",self,other)
        results = []
        for tab in self.tables:
            results.append(tab*other)
        lst = []
        for res in results:
            for tab in res.tables:
                lst.append(tab)
        #YoungTableaux.gene += 1
        return(self.__class__(lst,self.gen))

    def print(self,latex=False):
        print(self.dims)
        for d in self.dims:
            mults={tab for tab in self.tables if tab.dim_txt == d}
            unique_shapes = len(set(tuple(m.part_rows) for m in mults))
            if unique_shapes > 1:
                logger.info(f"Renaming ambiguous multiplets..")
                for m in mults:
                    #print(m.part_cols)
                    m.dim_txt += "_"+f"({','.join(str(x) for x in m.part_cols)})"
        counts = Counter([sol.dim_txt for sol in self.tables])
        pattern = r'(\d+)(?:b)?(?:_\(\d+(?:,\d+)*\))?'
        #sorted_tables = sorted(counts.items(), key=lambda item: ((lambda s: int(s[:-1]) if s.endswith('b') else int(s))(item[0])))
        logger.info(f"Sorting multiplets for printing...")
        sorted_tables = sorted(counts.items(), key=lambda item: ((lambda s: int(re.search(pattern, s).group(1)))(item[0])))
        print("Total number of tables:", len(self.tables))
        print("Decomposition:")
        if latex:
            latex_code=""
            normal=""
            for mult, count in sorted_tables:
                normal += "{0}*{1}+".format(count,mult)

                mult_name = re.search(r'(\d+)', mult).group(1)
                has_b = 'b' in mult
                subscript = re.search(r'_\((.*?)\)', mult)

                # Build the LaTeX string
                if has_b:
                    latex_code += f'{count}\\cdot\\overline{{{mult_name}}}'
                else:
                    latex_code += f'{count}\\cdot {mult_name}'

                if subscript:
                    latex_code += f'_{{({subscript.group(1)})}}'
                latex_code += r" \oplus "
            latex_code = (lambda s: s.rsplit(' ', 1)[0])(latex_code[:-1])
            print(normal[:-1])
            print("Latex source:")
            print(latex_code)
        else:
            normal=""
            for mult, count in sorted_tables:
                normal += "{0}*{1}+".format(count,mult)
            print(normal[:-1])

    def __iter__(self):
        return iter(self.tables)

class YoungTableau(object):
    @staticmethod
    def init_from_list(obj,Nc):
        while obj[-1] == 0:
            obj.remove(0)
        while obj[0] == Nc and len(obj)>1:
            obj.remove(Nc)
        part_rows = obj
        dims = (max(part_rows),len(part_rows))
        table = {}
        ncells = 0
        for row in range(dims[0]):
            for col in range(dims[1]):
                if part_rows[col] > row:
                    table[(row,col)] = 1
                    ncells += 1
                else:
                    table[(row,col)] = 0
        part_cols = list()
        for i in range(part_rows[0]):
            part_cols.append(sum(1 for j in part_rows if j>i))
        return (part_rows,part_cols,dims,ncells,table)

    def __init__(self, obj, Nc, parent1=None, parent2=None):
        self.Nc = Nc
        self.init_parents(parent1,parent2)
        #part_cols - (partition by colons) how many colons are in each row
        #part_rows - (partition by rows) how many rows are in each column
        if isinstance(obj,list):
            self.part_rows,self.part_cols,self.dims,self.ncells,self.table = self.init_from_list(obj,self.Nc)
        elif isinstance(obj,dict):
            nrows = list(obj)[-1][0]+1
            ncols = list(obj)[-1][1]+1
            part_rows = list()
            for col in range(ncols):
                n = 0
                for row in range(nrows):
                    if obj[(row,col)] != 0:
                        n = n+1
                part_rows.append(n)
            self.__init__(part_rows,self.Nc,parent1,parent2)
        self.dim = (self.calc_dim())
        conj_rows = list()
        for r in self.part_rows:
            conj_rows.append(self.Nc-r)
        conj_rows.append(self.Nc)
        conj_rows.reverse()
        conj_part_cols = self.init_from_list(conj_rows,self.Nc)[1]
        if len(self.part_cols)>len(conj_part_cols):
            self.dim_txt = str(self.dim)+"b"
        elif len(self.part_cols)==len(conj_part_cols) and self.part_cols[-1] > conj_part_cols[-1]:
            self.dim_txt = str(self.dim)+"b"
        else:
            self.dim_txt = str(self.dim)

    def init_parents(self,parent1,parent2):
        if parent1 is not None and parent1.parent1 is not None:
            self.parent1 = parent1
            self.parent2 = parent2
        else:
            self.parent1 = parent2
            self.parent2 = parent1

    def calc_dim(self):
        """
        Calculate the dimension of the multiplet
        """
        num = 1
        den = 1
        rows = self.dims[0]
        cols = self.dims[1]
        for i in range(rows):
          for j in range(cols):
            val = self.table[(i,j)]
            hooklen = 0
            for jj in range(j+1,cols):
              if self.table[(i,jj)] != 0: hooklen+=1
            for ii in range(i+1,rows):
              if self.table[(ii,j)] != 0: hooklen+=1
            if val  != 0:
              #print(val, end=' ')
              num *= self.Nc+j-i
              den *= hooklen+1
        return(int(num/den))

    def print(self):
        #print(self.part_cols)
        print(f"Y={self.part_rows} Dimension: {self.dim_txt}")
        for row in range(self.dims[0]):
            for col in range(self.dims[1]):
                #if self.table[col,row] != 0:
                print(self.table[row,col], end=' ')
            print()
        for col in range(self.dims[1]):
            print(u'\u2500', end=' ')
        print()

    def __iter__(self):
        self._index = 0
        return self

    def __next__(self):
        if self._index >= len(self.table):
            raise StopIteration
        else:
            item = list(self.table)[self._index]
            position = self._index+1
            self._index += 1
            if self.table[item] == 0:
                raise StopIteration
            return item, position

    def relabel(self):
        table = list()
        for row in range(self.dims[0]):
            for col in range(self.dims[1]):
                if self.table[row,col] != 0:
                    table.append(row*self.dims[1]+col+1)
        table.append(0)
        return(table)

    def id2labels(self,x):
        return((x-1)/self.dims[1]+1)

    def conjugate(self):
        conj_rows = list()
        for r in self.part_rows:
            conj_rows.append(self.Nc-r)
        conj_rows.append(self.Nc)
        conj_rows.reverse()
        return(self.__class__(conj_rows,self.Nc))

    def parent_list(self):
        parents = list()
        p = self.parent1
        while p != None:
            parents.append(p)
            p = p.parent1
        return(parents)

    def decompose(self):
        """
        Decomposition of a table in terms of quark and anti-quark projection
        operators as described in Appendix B of arXiv:1207.0609 + some
        optimization
        """
        logger.debug(f"Decomposing of {self.dim}-{self.part_rows} into q qbar tables")
        conj_rows = list()
        for r in self.part_rows:
            conj_rows.append(self.Nc-r)
        l = int(len(self.part_rows)/2)
        direction = 0 #Helps to avoid looping back and forth
        m=0
        n=0
        while l>0 and l<len(self.part_rows):
            m = sum(conj_rows[:l])
            n = sum(self.part_rows[l:])
            if m < n:
                new_direction = 1
                l += 1
            elif m > n:
                new_direction = -1
                l -= 1
            else:
                break
            if direction != 0 and direction != new_direction:
                break
            direction = new_direction

        if n==0:
            a = None
            b = None
            c = None
            logger.debug(f"Decomposition ended, firts occ: {n}")
        elif m!=n:
            a = None
            b = None
            c = None
            n = None
            print("This multiplet never appears in the product of adjoint representations.")
        else:
            l = len(self.part_rows)
            a = self.__class__(self.part_rows[:l-n],self.Nc)
            b = self.__class__(self.part_rows[l-n:],self.Nc)
            c=[m for m in a*b if m.dim != self.dim]
            logger.debug(f"Decomposition ended, {self.dim}={a.dim}x{b.dim} firts occ: {n}")
        self.first_occ = n
        #a and b are the antiquark and quark tables
        #c contains all the projectors which have to be subtracted
        self.decomposition = (a,b,c)

    def __copy__(self):
        copy_ = type(self).__new__(self.__class__)
        copy_.__dict__.update(self.__dict__)
        copy_.parent1 = None
        copy_.parent2 = None
        return copy_

    def __mul__(self,other):
        if self.dim < other.dim:
            return(other.__mul__(self))
        sol = database.get((tuple(self.part_cols),tuple(other.part_cols)))
        if sol:
            #sol = pickle.loads(pickle.dumps(sol,-1))
            #sol = copy.deepcopy(sol)
            for s in sol:
                s = copy.copy(s)
                s.init_parents(self,other)
            return(YoungTableaux(sol))
        block_num=99
        x = {}
        y = {}
        sol = z3.Solver()
        other_idx = other.relabel()
        rows = self.dims[0]+other.ncells
        cols = self.dims[1]+other.part_cols[0]
        if rows > self.Nc:
            rows = self.Nc
        #print("multing",self.dim,other.dim)
        #print(rows,cols)
        for row in range(rows):
            for col in range(cols):
                x[(row, col)] = z3.Int("x(%i,%i)" % (row, col))
                y[(row, col)] = z3.Int("y(%i,%i)" % (row, col))
                if row in range(self.dims[0]) and col in range(self.dims[1]):
                    if self.table[(row,col)] != 0:
                        sol.add(x[(row,col)] == block_num)
                    else:
                        #sol.add(z3.Or(x[(col, row)] == 1, x[(col, row)] == 0 ))
                        member_of(sol,x[(row,col)],other_idx)
                else:
                        #sol.add(z3.Or(x[(col, row)] == 1, x[(col, row)] == 0 ))
                        member_of(sol,x[(row,col)],other_idx)
                if col>0:
                    sol.add(z3.If(z3.And(x[(row, col-1)]==0, x[(row,col)]!=0),False,True))
                    sol.add(z3.If(z3.And(x[row,col-1]!=block_num,x[row,col-1]!=0,x[row,col]!=block_num,x[row,col]!=0),z3.If(x[row,col]>x[row,col-1],True,False),True))
                if row>0:
                    sol.add(z3.If(z3.And(x[(row-1,col)]==0, x[(row,col)]!=0),False,True))

        for row in range(rows):
            for col in range(cols):
                for row_i in range(rows):
                    for col_i in range(cols):
                        if row_i != row or col_i != col:
                            sol.add(z3.If(z3.And(x[row,col]!=block_num,x[row,col]!=0),x[(row, col)] != x[(row_i, col_i)],True))
                #Rewrite id of each cell using a,b notation (a=1,b=2,c=3,...)
                sol.add(z3.If(z3.And(x[row,col]!=block_num,x[row,col]!=0),y[(row, col)] == other.id2labels(x[(row,col)]),y[(row, col)] == x[(row, col)]))

        y_flat = [y[(row,col)] for row in range(rows) for col in reversed(range(cols))]
        repetitions = [z3.Int("p%i" % i) for i in range(other.part_rows[0])]
        #Count a,b label
        for iy,yy in enumerate(y_flat):
            for lab in range(other.part_rows[0]):
                repetitions[lab]=z3.Sum([z3.If(y_flat[i] == lab+1, 1,0) for i in range(iy+1)])
            #Reject If more b than a
            for lab1 in range(other.part_rows[0]):
                for lab2 in range(lab1,other.part_rows[0]):
                    if lab1 == lab2: continue
                    sol.add(repetitions[lab1]>=repetitions[lab2])
        #No repeating labels (a,b,c,...) in the column
        for lab in range(other.part_rows[0]):
            for col in range(cols):
                sol.add(z3.Sum([z3.If(y[(row,col)] == lab+1, 1,0) for row in range(rows)])<=1)

        for col in range(cols):
            for row in range(1,rows):
                sol.add(z3.If(z3.And(y[(row-1,col)]!=block_num,x[row,col]!=0,y[(row,col)]<y[(row-1,col)]),False,True))

        #Total num of cells = ncells in left table + ncells in right table
        sol.add(z3.Sum([z3.If(x[row,col] != 0, 1,0) for row in range(rows) for col in range(cols)]) == self.ncells + other.ncells )

        #Remove cols with zeroes
        #for row in range(rows):
            #z3.If(z3.Sum([x[(col,row)] for col in range (cols)])==0,x[(2,3)]==1,x[(2,3)]==1)

        #print(sol)
        z3.set_param("parallel.enable", True)
        #z3.set_param('parallel.threads.max', 4)
        num_solutions = 0
        solutions = list()
        while sol.check() == z3.sat:
            num_solutions += 1
            mod = sol.model()
            table = {}
            for row in range(rows):
                for col in range(cols):
                    table[(row, col)] = mod.eval(y[(row,col)]).as_long()
                    if table[(row,col)] == block_num:
                        table[(row,col)] = 9
            #remove_cols = 0
            #for row in range(rows):
                #lst=[table[(col, row)] for col in range(cols) if table[(col,row)]==0]
                #if not lst:
                    #remove_cols = row
#
            #ttable = {}
            #for col in range(cols):
                #for row in range(remove_cols+1,rows):
                    #ttable[(col,row-remove_cols-1)] = table[(col,row)]

            solutions.append(self.__class__(table,self.Nc,self,other))
            getDifferentSolutionMatrix(sol,mod,y,rows,cols)
            #break
        logger.debug(f"# Solutions: {num_solutions}")
        database[(tuple(self.part_cols),tuple(other.part_cols))] = solutions
        return(YoungTableaux(solutions))



import sys
import os
import logging

from .YoungTools import YoungTableau
from .YoungTools import YoungTableaux
from .Projectors import Projectors
from .utils import *

__all__ = [
       'YoungTableau',
       'YoungTableaux',
       'Projectors',
        ]



import z3

__all__ = [
       'getNewId',
       'getDifferentSolution',
       'getDifferentSolutionMatrix',
       'member_of',
       'generate_pairings',
        ]

def getNewId():
  return uuid.uuid4().int

def getDifferentSolution(sol,mod, *params):
  for t in params:
    sol.add(Or([t[i] != mod.eval(t[i]) for i in range(len(t))]))

# special case for a matrix; requires number of rows and columns
def getDifferentSolutionMatrix(sol,mod, x, rows, cols):
    sol.add(z3.Or([x[i,j] != mod.eval(x[i,j]) for i in range(rows) for j in range(cols)]))

def member_of(sol, e, v):
    sol.add(z3.Or([e == i for i in v]))

def generate_pairings(perms, nconns):
    """
    Create all possible connections of indices.
    Primaraly designed to fill in the treapezoid in Eq. (4.13) of
    arXiv:1207.0609.

    Here is a recursive version, which is more readable, but less efficient:
    def generate_pairings(perms,res,i):
        if not len(perms):
            if i == 3:
                print(res)
            return
        for j,perm in enumerate(perms):
            remaining = [p for p in perms[j+1:] if ((p[0]!=perm[0] and p[0]!=perm[1]) and (p[1]!=perm[0] and p[1]!=perm[1]))]
            res[i] = perm
            generate_pairings(remaining,res,i+1)
    """
    stack = [(perms, [], 0)]
    result = list()

    while stack:
        current_perms, res, i = stack.pop()

        if not len(current_perms):
            if i == nconns:
                result.append(res)
            continue

        for j, perm in enumerate(current_perms):
            remaining = [p for p in current_perms[j+1:] if ((p[0]!=perm[0] and p[0]!=perm[1]) and (p[1]!=perm[0] and p[1]!=perm[1]))]
            new_res = res + [perm]
            stack.append((remaining, new_res, i+1))
    return result
